/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Class
    primitiveMesh

Description
    Efficient cell-centre calculation using face-addressing, face-centres and
    face-areas.
    Cell centres and volumes for spherical geometry

\*---------------------------------------------------------------------------*/

#include "primitiveMesh.H"
//#include "IFstream.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// * * * * * * * * * * * * * Private Member Functions  * * * * * * * * * * * //

void primitiveMesh::calcCellCentresAndVols() const
{
    //if (debug)
    {
        Pout<< "primitiveMesh::calcCellCentresAndVols() : "
            << "Calculating cell centres and cell volumes"
            << " for sperical geometry"
            << endl;
    }

    // It is an error to attempt to recalculate cellCentres
    // if the pointer is already set
    if (cellCentresPtr_ || cellVolumesPtr_)
    {
        FatalErrorIn("primitiveMesh::calcCellCentresAndVols() const")
            << "Cell centres or cell volumes already calculated"
            << exit(FatalError);
    }

    // set the accumulated cell centre to zero vector
    cellCentresPtr_ = new vectorField(nCells());
    vectorField& cellCtrs = *cellCentresPtr_;

    // Initialise cell volumes to 0
    cellVolumesPtr_ = new scalarField(nCells());
    scalarField& cellVols = *cellVolumesPtr_;

    // Make centres and volumes
    makeCellCentresAndVols(faceCentres(), faceAreas(), cellCtrs, cellVols);

    if (debug)
    {
        Pout<< "primitiveMesh::calcCellCentresAndVols() : "
            << "Finished calculating cell centres and cell volumes"
            << endl;
    }
}


void primitiveMesh::makeCellCentresAndVols
(
    const vectorField& fCtrs,
    const vectorField& fAreas,
    vectorField& cellCtrs,
    scalarField& cellVols
) const
{
    // Clear the fields for accumulation
    cellCtrs = vector::zero;
    cellVols = 0.0;
    
    const labelList& own = faceOwner();
    const labelList& nei = faceNeighbour();

    // next calculate exact cell volume and centre for spherical geometry

    scalarField r1(nCells(), scalar(0));
    scalarField r2(nCells(), scalar(0));
    scalarField r(nCells(), scalar(0));
    scalarField omega(nCells(), scalar(0));      // solid angle of cell
    
    forAll (faces(), faceI)
    {
        scalar rF = mag(fCtrs[faceI]);
        scalar Af = mag(fAreas[faceI]);
        scalar cosCfAf = (fAreas[faceI] & fCtrs[faceI])/(rF*Af);
        
        // determine whether radial or spherical face
        
        if(mag(cosCfAf) < SMALL)     // radial face
        {
            if(mag(r[own[faceI]]) < SMALL) r[own[faceI]] = rF;
            if(mag(r[nei[faceI]]) < SMALL) r[nei[faceI]] = rF;
            
            if(mag(r[own[faceI]] - rF) > 100*rF*SMALL)
            {
                FatalErrorIn("primitiveMesh::makeCellCentresAndVols")
                    << " for cell " << own[faceI] 
                    << " not all radial faces have the same radius, "
                    << "mag(fCtrs[" << faceI << " ]) = " 
                    << mag(fCtrs[faceI]) << " but previous radius = " 
                    << r[own[faceI]]<< exit(FatalError);
            }

            if(mag(r[nei[faceI]] - rF) > 100*rF*SMALL)
            {
                FatalErrorIn("primitiveMesh::makeCellCentresAndVols")
                    << " for cell " << nei[faceI] 
                    << " not all radial faces have the same radius, "
                    << "mag(fCtrs[" << faceI << " ]) = " 
                    << mag(fCtrs[faceI]) << " but previous radius = " 
                    << r[nei[faceI]] << exit(FatalError);
            }
        }
        
        else if(mag(mag(cosCfAf) - scalar(1)) < SMALL) // sph face
        {
            // set the radii of the owner cell
            if(mag(r1[own[faceI]]) < SMALL)
            {
                r1[own[faceI]] = rF;
            }
            else if
            (
                mag(r2[own[faceI]]) < SMALL 
                && mag(rF - r1[own[faceI]]) > SMALL
            )
            {
                r2[own[faceI]] = rF;
            }
            else if
            (
                mag(r2[own[faceI]]) > SMALL 
                && mag(rF - r1[own[faceI]]) > SMALL
                && mag(rF - r2[own[faceI]]) > SMALL
            )
            {
                FatalErrorIn("primitiveMesh::makeCellCentresAndVols")
                    << " for cell " << own[faceI] << " face " << faceI 
                    << " has >2 spherical faces\n"
                    << " r1[own[faceI]] = " << r1[own[faceI]] 
                    << " r2[own[faceI]] = " << r2[own[faceI]]
                    << " rF = " << rF
                    << exit(FatalError);       
            }

            // set the radii of the neighbour cell
            if (isInternalFace(faceI))
            {
                if(mag(r1[nei[faceI]]) < SMALL)
                {
                    r1[nei[faceI]] = rF;
                }
                else if
                (
                    mag(r2[nei[faceI]]) < SMALL 
                    && mag(rF - r1[nei[faceI]]) > SMALL
                )
                {
                    r2[nei[faceI]] = rF;
                }
                else if
                (
                    mag(r2[nei[faceI]]) > SMALL 
                    && mag(rF - r1[nei[faceI]]) > SMALL
                    && mag(rF - r2[nei[faceI]]) > SMALL
                )
                {
                    FatalErrorIn("primitiveMesh::makeCellCentresAndVols")
                        << " for cell " << nei[faceI] << " face " << faceI
                        << " has >2 spherical faces\n" 
                        << " r1[nei[faceI]] = " << r1[nei[faceI]] 
                        << " r2[nei[faceI]] = " << r2[nei[faceI]]
                        << " rF = " << rF
                        << exit(FatalError);
                }
            }
            
            // set the solid angle of the owner cell
            scalar om = Af/sqr(rF);
            if(mag(omega[own[faceI]]) < SMALL)
            {
                omega[own[faceI]] = om;
            }
            else if(mag(omega[own[faceI]] - om) > SMALL)
            {
                FatalErrorIn("primitiveMesh::makeCellCentresAndVols")
                    << " for cell " << own[faceI] 
                    << ", spherical faces have 2 different solid angles "
                    << omega[own[faceI]] << " and " << om
                    << exit(FatalError);
            }

            // set the solid angle of the neighbour cell
            if (isInternalFace(faceI))
            {
                if(mag(omega[nei[faceI]]) < SMALL)
                {
                    omega[nei[faceI]] = om;
                }
                else if(mag(omega[nei[faceI]] - om) > SMALL)
                {
                    FatalErrorIn("primitiveMesh::makeCellCentresAndVols")
                        << " for cell " << nei[faceI] 
                        << ", spherical faces have 2 different solid angles "
                        << omega[nei[faceI]] << " and " << om
                        << exit(FatalError);
                }
            }
            
            // set the cell centre direction for the owner cell
            vector rHat = fCtrs[faceI]/mag(fCtrs[faceI]);
            if(mag(cellCtrs[own[faceI]]) < SMALL) cellCtrs[own[faceI]] = rHat;
            else if(mag((rHat & cellCtrs[own[faceI]]) - scalar(1)) > SMALL)
            {
                FatalErrorIn("primitiveMesh::makeCellCentresAndVols")
                    << " for cell " << own[faceI] 
                    << ", spherical faces have 2 different directions "
                    << cellCtrs[own[faceI]] << " and " << rHat
                    << exit(FatalError);
            }
            
            // set the cell centre direction for the neighbour cell
            if (isInternalFace(faceI))
            {
                if(mag(cellCtrs[nei[faceI]]) < SMALL)
                {
                    cellCtrs[nei[faceI]] = rHat;
                }
                else if(mag((rHat & cellCtrs[nei[faceI]]) - scalar(1)) > SMALL)
                {
                    FatalErrorIn("primitiveMesh::makeCellCentresAndVols")
                        << " for cell " << nei[faceI] 
                        << ", spherical faces have 2 different directions "
                        << cellCtrs[nei[faceI]] << " and " << rHat
                        << exit(FatalError);
                }
            }
        }
        else // error - neither spherical nor radial face
        {
            FatalErrorIn("primitiveMesh::makeCellCentresAndVols")
                << " face " << faceI 
                << " is neither a spherical nor radial face since "
                << "Cf.Af/(|Cf||Af|) = " << cosCfAf 
                << exit(FatalError);
        }
    }

    cellCtrs *= r;
    cellVols = mag((1./3.)*omega*(pow(r2,3) - pow(r1,3)));
}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

const vectorField& primitiveMesh::cellCentres() const
{
    if (!cellCentresPtr_)
    {
        calcCellCentresAndVols();
    }

    return *cellCentresPtr_;
}


const scalarField& primitiveMesh::cellVolumes() const
{
    if (!cellVolumesPtr_)
    {
        calcCellCentresAndVols();
    }

    return *cellVolumesPtr_;
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// ************************************************************************* //
